function EKF_localization()
%% init
    close all 
    clear all

    set(0,'defaultTextFontName', 'Arial');
    set(0,'defaultTextFontSize', 14);
    set(0,'defaultTextFontWeight', 'Bold');
    set(0,'defaultTextBackgroundColor', [1,1,1]);
    set(0,'defaultLineLineWidth', 3);
    
%% execution flags
    flags.skip_meas   = false; % if true the measurement step will be skipped (only prediction will be performed)
    flags.skip_update = false; % if true the update step will be skipped (only prediction and measurement prediction)
    flags.animation = true; % if false no animation
    flags.amimation_pause = false; % if true pause (key press) at each iteration, otherwise small timed pause occurs
    flags.kNoise = 1; % set to 0 to avoid random noise addition in initial position and prediction
    flags.kNoiseMeas = 1; %set to 0 to avoid random noise addition in measurement
    
    drawaxisvalues = [-3 3 -4 3]; % axis limits for the figure

%% robot
    % robot, used as inputs for the system
    robot.dt = 0.1;
    robot.baseSpeed = 0.5; % [m/s], used to generate motion

    % information used for robot drawings
    robot.shape2D = [
        .6       -.3       -.3       .6
        0       0.25     -0.25    0
        1       1       1       1
        ];
    robot.color = [0 .7 0];

    % robot complete motion
    robot.vw_inputs = [ ...
        rectMovementBaseSpeed(.5, robot.baseSpeed,robot.dt); ...
        arcMovementBaseSpeed(.5, 2*pi, robot.baseSpeed , robot.dt); ...
        arcMovementBaseSpeed(.5,  - pi, robot.baseSpeed , robot.dt); ...
        rectMovementBaseSpeed(1.5, robot.baseSpeed,robot.dt); ...
        arcMovementBaseSpeed(1, - pi, robot.baseSpeed , robot.dt); ...
        rectMovementBaseSpeed(.5, robot.baseSpeed,robot.dt); ...
        arcMovementBaseSpeed(.5, - pi / 2, robot.baseSpeed , robot.dt); ...
        rectMovementBaseSpeed(2.5, robot.baseSpeed,robot.dt); ...
        arcMovementBaseSpeed(.5, - pi, robot.baseSpeed , robot.dt); ...
        rectMovementBaseSpeed(1.5, robot.baseSpeed, robot.dt); ...
        arcMovementBaseSpeed(.5, - pi / 2, robot.baseSpeed , robot.dt); ...
        rectMovementBaseSpeed(1.0, robot.baseSpeed, robot.dt); ...
    ];
    % the delta motions are in x and theta given the velocities and sample
    % time
    robot.delta_motion = [robot.vw_inputs(:,1) * robot.dt, zeros(size(robot.vw_inputs,1),1), robot.vw_inputs(:,2) * robot.dt];
    robot.time = (0:size(robot.vw_inputs, 1)) * robot.dt;
    
    tid = 1; %iteration counter


    % robot position start
    robot.init_state = [0 0 0]; % robot starting point (x, y, theta)
    % prepare storage space to save all the robot trajectory
    robot.pose = nan(length(robot.time),3);
    robot.pose(1,:) = [robot.init_state];
    
%% sensor information

% minimum measured distance
    sensor.rhomin = .3;

% uncomment one of the nexts sensor definition

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% worst sensor, very bad localization
%     sensor.rhomax = 1.;
%     sensor.anglemin = -pi/12;
%     sensor.anglemax = +pi/12;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% worse sensor
    sensor.rhomax = 1.5;
    sensor.anglemin = -pi/6;
    sensor.anglemax = +pi/6;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% good sensor
%     sensor.rhomax = 2.1;
%     sensor.anglemin = -pi/4;
%     sensor.anglemax = +pi/4;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% % best sensor
%     sensor.rhomax = 2.5;
%     sensor.anglemin = -pi/2;
%     sensor.anglemax = +pi/2;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% map of the environment
    
    % setup the map
    [map.x, map.y] = meshgrid(-2:2,-3:2);
    map.x = reshape(map.x, size(map.x,1)*size(map.x,2),1);
    map.y = reshape(map.y, size(map.y,1)*size(map.y,2),1);
    for i = 1:length(map.x)
        map.names{i} = sprintf('%d', i);
    end
    
%% setup the EKF 
    R = diag([.03,.03,deg2rad(3)])^2; % noise on delta x, delta y, delta theta in motion
    Rstd = sqrt(R);
    Q = diag([.03, deg2rad(3)])^2; % noise on rho and theta in measurement
%     Q = diag([.3, deg2rad(30)])^2; % very bad measurement system (high error variance)
    Qstd = sqrt(Q);
    
    % init state of the filter, i.e., initial mean and covariance
    ekf.sigma = diag([.1,.1,deg2rad(10)])^2;
    ekf.mu = robot.init_state';
    
    % add random to true robot position
    robot.pose(1,:) = robot.pose(1,:)  + (sqrt(ekf.sigma) * randn(3,1) * flags.kNoise)';
    
    % save ekf history
    ekfHist{1} = ekf;
    
    % create storage for trajectory and trajectory variance values
    poseEst = nan(length(robot.time),3);
    poseEst(1,:) = ekf.mu';
    stdEst  = nan(length(robot.time),3);
    stdEst(1,:)  = diag(sqrt(ekf.sigma));
    
    % plot first animation figure
    if(flags.animation)
        figworld = figure(); % figure with robot in the world
        clf,hold on,grid on,box on; % clear figure and setup plots
        if(~flags.skip_meas) % if there is mesaurement process
            drawRobot(robot.pose(tid,:),robot.shape2D,'b', sensor); % draw true robot and sensor
            drawRobot(ekf.mu ,robot.shape2D,'r'); % draw the estimated robot
            error_ellipse('C',ekf.sigma(1:2,1:2), 'mu',ekf.mu, 'conf', .95, 'style', 'r'); % draw the error ellipse on robot
            plot(map.x, map.y, 'm*'); % draw map and write map names
            text(map.x + 0.2, map.y + 0.2, map.names);
        else
            drawRobot(robot.pose(tid,:),robot.shape2D,'b'); % draw true robot
            drawRobot(ekf.mu ,robot.shape2D,'r'); % draw estimated robot
            error_ellipse('C',ekf.sigma(1:2,1:2), 'mu',ekf.mu, 'conf', .95, 'style', 'r'); % draw the error ellipse on robot
        end
        axis equal,axis(drawaxisvalues),hold off;
        drawnow()
        
        if(~flags.skip_meas)
            figrobot = figure(); %se tup a figure for the local robot drawings
            clf;hold on,grid on,box on;
            drawRobot([0,0,0] ,robot.shape2D,'r',sensor); % draw robot and sensor
            axis equal,hold off,drawnow();
        end
        
        fprintf('press a key to continue...');
        pause
        fprintf('OK\n');
    end
    

%% execution cycle
    for tid = 2 : length(robot.time)
        % evolute true robot system
        robot.pose(tid,:) = nominalMotion(robot.pose(tid-1,:) , robot.delta_motion(tid-1,:) );
        % add some random noise to the robot motion
        input_noised = robot.delta_motion(tid-1,:)' + Rstd * randn(3,1) * flags.kNoise;
        % evolute ekf system (robot position estimation by localization)
        ekf = EKF_predict(ekf, input_noised, R);
        
        if(~flags.skip_meas)            
            % get measurement from sensor
            [index_measure_map, measures] = measureMap(sensor, robot.pose(tid,:), map);
            % measures are organized in a matrix, first column is rho, second is theta
            for i = 1 : size(measures,1)
                % add random to the real measurement
                measures(i,:) = (measures(i,:)' + Qstd * randn(2,1) * flags.kNoiseMeas)';
            end
            
            % measure from state (i.e., measurement prediction)
            % I will force index that are measured from sensor but not
            % predictable by state
            [index_measure_state, measures_state] = measureMap(sensor, ekf.mu, map, index_measure_map);
            
            % for each measurement available (i.e., intersection between
            % measurement from sensor and from state)
            % collect measurement and do update
            index_measure_update = intersect(index_measure_state, index_measure_map);
            h = []; z = []; H = [];  M = []; Q_meas = []; S_meas_pred = {};
            for index = index_measure_update
                % get the predicted measure
                h_1 = measures_state(index_measure_state == index,:)';
                % get the measure value
                z_1 = measures(index_measure_map == index,:)';
                % calculate jacobian of the measure
                [htmp, H_1, M_1, S] = EKF_jacobian_measurePrediction(ekf, [map.x(index), map.y(index)], Q);
                % check that there are no implementation errors
                assert( sum((htmp - h_1).^2) < 1e-15 );
                % collect measurement prediction covariance, will be used
                % to plot ellipses
                S_meas_pred{length(S_meas_pred) + 1} = S;
                
                % stack measurement information (h, z, H, M, Q_meas)
                h = [h; h_1]; H = [H; H_1]; z = [z; z_1];
                M = [M,                             zeros(size(M,1),size(M_1,2));
                     zeros(size(M_1,1),size(M,2)),  M_1                          ];
                Q_meas = [  Q_meas,                             zeros(size(Q_meas,1),size(Q,2));
                            zeros(size(Q,1),size(Q_meas,2)),    Q                               ];
                 
            end
            
            if(~flags.skip_update)   && (~isempty(index_measure_update))
                [ekf] = ekf_update(ekf,h,z,H,M,Q_meas); % perform update
            end
            
        else
            index_measure_map = [];
            measures = [];
        end
        
        % save last estimation and std dev values
        poseEst(tid,:) = ekf.mu';
        stdEst(tid,:)  = diag(sqrt(ekf.sigma))';

        if(flags.animation)
            figure(figworld) % get the figure
            clf; hold on; grid on; box on; % clear
            plot(robot.pose(1:tid,1),robot.pose(1:tid,2),'b'); % plot true trajectory
            plot(poseEst(1:tid,1),poseEst(1:tid,2),'r'); % plot estimated trajectory
            if(~flags.skip_meas)
                drawRobot(robot.pose(tid,:),robot.shape2D,'b', sensor); % draw true robot
                drawRobot(ekf.mu ,robot.shape2D,'r'); % draw estimated robot
                error_ellipse('C',ekf.sigma(1:2,1:2), 'mu',ekf.mu, 'conf', .95, 'style', 'r'); % draw estimated covariance
                plot(map.x, map.y, 'm*'); % draw map and labels
                text(map.x + 0.2, map.y + 0.2, map.names);
            else
                drawRobot(robot.pose(tid,:),robot.shape2D,'b');; % draw true robot
                drawRobot(ekf.mu ,robot.shape2D,'r'); % draw estimated robot
                error_ellipse('C',ekf.sigma(1:2,1:2), 'mu',ekf.mu, 'conf', .95, 'style', 'r'); % draw estimated covariance
            end
            % measured features
            
            if(~flags.skip_meas) && (~isempty(index_measure_update))
            
                plot(map.x(index_measure_map), map.y(index_measure_map), 'g*'); % put a green star on the measured points of the map
                T_W_R = [   cos(robot.pose(tid,3)), -sin(robot.pose(tid,3)), robot.pose(tid,1); 
                            sin(robot.pose(tid,3)),  cos(robot.pose(tid,3)), robot.pose(tid,2);
                                0,              0,          1]; % pose of the true robot wrt world
                [meas_cart_x, meas_cart_y] = pol2cart(measures(:,2),measures(:,1)); % convert measurement values in cartesian coordinate (measure are in robot frame)
                meas = T_W_R * [meas_cart_x'; meas_cart_y'; ones(1,length(meas_cart_x))]; % rotate and traslate in world reference
                plot(meas(1,:), meas(2,:), 'bo'); % plot the measurement of the sensor in world coordinate

                % measurement prediction by ekf state
                [meas_cart_x_pred, meas_cart_y_pred] = pol2cart(measures_state(:,2),measures_state(:,1)); % convert predicted measurement values in cartesian coordinate (measure are in robot frame)
            end
            axis equal;axis(drawaxisvalues);hold off;drawnow();
            
            if(~flags.skip_meas) 
                figure(figrobot) % plot the robot figure
                clf;hold on;grid on;box on;
                drawRobot([0,0,0] ,robot.shape2D,'r',sensor); % draw the robot
                if(~isempty(index_measure_update))
                    plot(meas_cart_x_pred, meas_cart_y_pred, 'co'); % draw the predicted measures
                    plot(meas_cart_x, meas_cart_y, 'bo'); % draw the sensor measures

                    % select points of the map that are measured
                    real_points = [map.x(union(index_measure_map,index_measure_state))'; map.y(union(index_measure_map,index_measure_state))'; ones(1,length(union(index_measure_map,index_measure_state)))];
                    
                    % rototranslate in robot frame
                    real_points = T_W_R^(-1) * real_points;
                    % plot map point in robot frame
                    plot(real_points(1,:), real_points(2,:), 'g*');
                    
                    % put labels on map elements
                    u = union(index_measure_map,index_measure_state);
                    s = {};
                    for i = 1 : length(union(index_measure_map,index_measure_state))
                        s{i} = sprintf('%d',u(i));
                    end
                    text(real_points(1,:) + .1, real_points(2,:) + .1, s);
                    s = {};
                    for i = 1 : length(index_measure_state)
                        s{i} = sprintf('%d',index_measure_state(i));
                    end
                    text(meas_cart_x_pred + .1, meas_cart_y_pred + .1, s);

                    % draw error ellipses of the measures
                    for i = 1 : length(S_meas_pred)
                        [xx,yy] = pol2cart(h( (i-1) * 2 + 2), h( (i-1) * 2 + 1));
                        error_ellipse('C',S_meas_pred{i}, 'mu', [xx,yy], 'conf', .95, 'style', 'c');
                    end
                end

                axis equal,hold off,drawnow();
            end
            
            if(flags.amimation_pause)
                fprintf('press a key to continue...');
                pause
                fprintf('OK\n');
            else
                pause(.01)
            end

        end
        
    end
    
    fprintf('press a key to continue...');
    pause
    fprintf('OK\n');
    
    robot.pose(:,3) = unwrap(robot.pose(:,3));
    poseEst(:,3) = unwrap(poseEst(:,3));
    
    figure()
    hold on
    plot(robot.pose(:,1),robot.pose(:,2),'b')
    plot(poseEst(:,1),poseEst(:,2),'r')
    hold off

    figure()
    subplot(3,1,1)
    hold on
    plot(robot.time,robot.pose(:,1),'b')
    plot(robot.time,poseEst(:,1),'r')
    plot(robot.time,poseEst(:,1) + 3 * stdEst(:,1), 'r--');
    plot(robot.time,poseEst(:,1) - 3 * stdEst(:,1), 'r--');
    hold off
    subplot(3,1,2)
    hold on
    plot(robot.time,robot.pose(:,2),'b')
    plot(robot.time,poseEst(:,2),'r')
    plot(robot.time,poseEst(:,2) + 3 * stdEst(:,2), 'r--');
    plot(robot.time,poseEst(:,2) - 3 * stdEst(:,2), 'r--');
    hold off
    subplot(3,1,3)
    hold on
    plot(robot.time,robot.pose(:,3),'b')
    plot(robot.time,poseEst(:,3),'r')
    plot(robot.time,poseEst(:,3) + 3 * stdEst(:,3), 'r--');
    plot(robot.time,poseEst(:,3) - 3 * stdEst(:,3), 'r--');
    hold off

    

end

function [selectedIndex, measures] = measureMap(sensor, pose, map, index_measure_map)
    % get the distance and angle of points of the map in the robot
    % reference frame (i.e., perform measures)
    T_W_R = [   cos(pose(3)), -sin(pose(3)), pose(1);
                sin(pose(3)),  cos(pose(3)), pose(2);
                    0,              0,          1];
    T_R_W = [T_W_R(1:2,1:2)', - T_W_R(1:2,1:2)' * T_W_R(1:2,3); 0 0 1];

    map_R = T_R_W * [map.x'; map.y'; ones(1,length(map.x))];
    
    % these are measures for all the map
    map_rho   = sqrt(map_R(1,:).^2 + map_R(2,:).^2);
    map_theta = atan2(map_R(2,:), map_R(1,:));
    
    % save a copy of the measures
    map_rho_orig = map_rho;
    map_theta_orig = map_theta;
    selectedIndex = 1:length(map_rho);
    
    % select measurements that respect physical sensor
    indexes   = find(map_rho <= sensor.rhomax);
    map_rho   = map_rho(indexes);
    map_theta = map_theta(indexes);
    selectedIndex = selectedIndex(indexes);
    
    indexes   = find(map_rho >= sensor.rhomin);
    map_rho   = map_rho(indexes);
    map_theta = map_theta(indexes);
    selectedIndex = selectedIndex(indexes);

    indexes   = find(map_theta <= sensor.anglemax);
    map_rho   = map_rho(indexes);
    map_theta = map_theta(indexes);
    selectedIndex = selectedIndex(indexes);

    indexes   = find(map_theta >= sensor.anglemin);
    map_rho   = map_rho(indexes);
    map_theta = map_theta(indexes);
    selectedIndex = selectedIndex(indexes);
    
    if(nargin == 4)
        % select elements that are not selected by the measurement process
        % but are "forced" by the additional input (i.e., the true sensor
        % has measured them, but they are invalid for the prediction)
        force_index = setdiff(index_measure_map,selectedIndex);
        selectedIndex = [selectedIndex, force_index];
        map_rho   = [map_rho map_rho_orig(force_index)];
        map_theta = [map_theta map_theta_orig(force_index)];
    end

    measures = [map_rho', map_theta'];
end

function [ekf] = ekf_update(ekf,h,z,H,M,Q)
    % perform standard EKF update
    S = H * ekf.sigma * H' + M * Q * M';
    invS = S^(-1);
    K = ekf.sigma * H' * invS;
    ekf.mu = ekf.mu + K * (z-h);
    ekf.sigma = (eye(length(ekf.mu)) - K * H) * ekf.sigma;
end


function [h, H, M, S] = EKF_jacobian_measurePrediction(ekf, map_point, Q)
    % values and jacobian of the measurement prediction
    x  = ekf.mu(1);
    y  = ekf.mu(2);
    th = ekf.mu(3);
    
    px = map_point(1);
    py = map_point(2);
    
    rho = ...
    ((px*cos(th) - x*cos(th) + py*sin(th) - y*sin(th))^2 + (py*cos(th) - y*cos(th) - px*sin(th) + x*sin(th))^2)^(1/2);


    theta = ...
    atan2((py*cos(th) - y*cos(th) - px*sin(th) + x*sin(th)),(px*cos(th) - x*cos(th) + py*sin(th) - y*sin(th)));


    H =[
                                              -(2*cos(th)*(px*cos(th) - x*cos(th) + py*sin(th) - y*sin(th)) - 2*sin(th)*(py*cos(th) - y*cos(th) - px*sin(th) + x*sin(th)))/(2*((px*cos(th) - x*cos(th) + py*sin(th) - y*sin(th))^2 + (py*cos(th) - y*cos(th) - px*sin(th) + x*sin(th))^2)^(1/2)),                                           -(2*cos(th)*(py*cos(th) - y*cos(th) - px*sin(th) + x*sin(th)) + 2*sin(th)*(px*cos(th) - x*cos(th) + py*sin(th) - y*sin(th)))/(2*((px*cos(th) - x*cos(th) + py*sin(th) - y*sin(th))^2 + (py*cos(th) - y*cos(th) - px*sin(th) + x*sin(th))^2)^(1/2)),  0
     (sin(th)/(px*cos(th) - x*cos(th) + py*sin(th) - y*sin(th)) + (cos(th)*(py*cos(th) - y*cos(th) - px*sin(th) + x*sin(th)))/(px*cos(th) - x*cos(th) + py*sin(th) - y*sin(th))^2)/((py*cos(th) - y*cos(th) - px*sin(th) + x*sin(th))^2/(px*cos(th) - x*cos(th) + py*sin(th) - y*sin(th))^2 + 1), -(cos(th)/(px*cos(th) - x*cos(th) + py*sin(th) - y*sin(th)) - (sin(th)*(py*cos(th) - y*cos(th) - px*sin(th) + x*sin(th)))/(px*cos(th) - x*cos(th) + py*sin(th) - y*sin(th))^2)/((py*cos(th) - y*cos(th) - px*sin(th) + x*sin(th))^2/(px*cos(th) - x*cos(th) + py*sin(th) - y*sin(th))^2 + 1), -1];

    M =[  1     0
         0     1    ];
     
     h = [rho;theta];
     
     S = H * ekf.sigma * H' + M * Q * M';
     
end

function pose = nominalMotion(pose,u)
    % nominal motion
    x  = pose(1);
    y  = pose(2);
    th = pose(3);
    
    dx  = u(1);
    dy  = u(2);
    dth = u(3);
       
    x1  = cos(th)*(dx) - sin(th)*(dy) + x;
    y1  = sin(th)*(dx) + cos(th)*(dy) + y;
    th1 = th + dth;    
    
    pose = [x1,y1,th1];
end

function ekf = EKF_predict(ekf, u, R)
    % EKF prediction
    x  = ekf.mu(1);
    y  = ekf.mu(2);
    th = ekf.mu(3);

    dx  = u(1);
    dy  = u(2);
    dth = u(3);

    G = [ 1, 0, - dy*cos(th) - dx*sin(th);
         0, 1,   dx*cos(th) - dy*sin(th);
         0, 0,                         1];
     
    N = [ cos(th), -sin(th), 0;
          sin(th),  cos(th), 0;
           0,        0,      1];
       
    x1  = cos(th)*(dx) - sin(th)*(dy) + x;
    y1  = sin(th)*(dx) + cos(th)*(dy) + y;
    th1 = th + dth;
    
    ekf.mu = [x1,y1,th1]';
    ekf.sigma = G * ekf.sigma * G' + N * R * N';
end

function drawRobot(p,shape,color,sensor)
    % draw the robot shape and (optionally) the sensor range
    M = [cos(p(3)), -sin(p(3)), p(1);
         sin(p(3)),  cos(p(3)), p(2);
            0       0           1];
    pts = M * shape;
    fill(pts(1,:),pts(2,:),color);
    if(nargin == 4)
        % draw the sensor
        
        nl = 6;
        angle = linspace(sensor.anglemin,sensor.anglemax,nl);
        l1 = [cos(angle) * sensor.rhomin; sin(angle) * sensor.rhomin; ones(1,nl)];
        l2 = [cos(angle) * sensor.rhomax; sin(angle) * sensor.rhomax; ones(1,nl)];
        
%         pts = M * [l1, l2, l1(:,1)];
        pts = M * [l1 l2(:,end:-1:1) l1(:,1)];
        plot(pts(1,:), pts(2,:), 'g');
    end
end